//RJA trying to create this WINDOW class

#ifndef _CURSES_H
#define _CURSES_H

//RJA adapted this from https://github.com/mikeyk730/Rogue-Collection/blob/master/readme.md
//which is I believe based on pdcurses

//RJA:  Note had some trouble here because the original was in C++ with a window class that was too complex for FlexCC

//Notes by RJA (who is figuring out how curses works...)
//  curscr is the window whose data will by displayed in terminal
//  stdscr is the usual screen used for drawing, copied to curscr by refresh()
//  other windows can be created and then copied to curscr by wrefresh()

#include "Platform.h"  //P2 Spin settings and such

//choose input source
//#define USE_USB_KEYBOARD  //uses PC if not set

//define output options
//#define DISPLAY_IN_TERMINAL

#include "string.h"
#include "stdio.h"
#include "stdbool.h"
#include <stdarg.h>

// the 2bpp tile driver with VGA output
struct __using("1080p_TileDriver_8b.spin2") vga; //RJA:  Visual Studio will show an error here, but it's OK

// access to the cogless graphics driver inside the vga driver
struct __using("Graphics3g.spin2") gr; //RJA:  Visual Studio will show an error here, but it's OK

//GarryJ's one cog USB Keyboard/Mouse driver
struct __using("1CogKbM_rja3b.spin2") kbm; //RJA:  Visual Studio will show an error here, but it's OK


typedef struct
{
    int x;
    int y;
    int MaxX;
    int MaxY;
    int buttons;
    char scan0;
    char scan1;
    char scan2;
    char scan3;
    char scan4;
    char scan5;
    char scan6;
    char scan7;

} MouseInfo;

MouseInfo MouseStatus = { 0,0,0,0 };
MouseInfo OldMouseStatus = { 0,0,0,0 };

int DebugLong;


void StartMouse()
{//Start up the USB mouse
    printf("Starting Mouse...\n");
    MouseStatus.MaxX = 1280 - 1;
    MouseStatus.MaxY = 720 - 1;
    MouseStatus.x = 100;
    MouseStatus.y = 100;
    kbm.Start(USB_BASE_PIN, USB_ENABLE_PIN, USB_DM_PIN, USB_ERR_LED_PIN, &MouseStatus);  //(base_pin, dm_pin, err_led_pin, pMouseData) : cog
}

void InitUserInterface()
{//start P2 vga & mouse & files

    //change from default baud rate
    _setbaud(_BAUD);

    //Note:  Clock frequency set with enum in Platform.h
//check clock freq
    int cf = _clockfreq();
    printf("clock freq= %d\n", cf);

    printf("Starting Graphics...\n");
    int pCommand = gr.start();
    int pFont = gr.GetFont();

    //start USB mouse
    StartMouse();

    printf("Starting VGA...\n");
    vga.start(&MouseStatus, pCommand, pFont, &DebugLong);
    vga.cls(vga.MidnightBlue);

    //RJA got this ascii art from:  https://the-spoiler.com/ADVENTURE/Computer.systems/rogue.2.html
    vga.setforecolor(vga.White);
    vga.setbackcolor(vga.Temptress);
    vga.row = 0;
    vga.col = 31;
    char* rogue1;
    rogue1 = "     _______  _______  _______           _______     \r";
    vga.print_string2(rogue1);
    rogue1 = "    (  ____ )(  ___  )(  ____ \\|\\     /|(  ____ \\    \r";
    vga.print_string2(rogue1);
    rogue1 = "    | (    )|| (   ) || (    \\/| )   ( || (    \\/    \r";
    vga.print_string2(rogue1);
    rogue1 = "    | (____)|| |   | || |      | |   | || (__        \r";
    vga.print_string2(rogue1);
    rogue1 = "    |     __)| |   | || | ____ | |   | ||  __)       \r";
    vga.print_string2(rogue1);
    rogue1 = "    | (\\ (   | |   | || | \\_  )| |   | || (          \r";
    vga.print_string2(rogue1);
    rogue1 = "    | ) \\ \\__| (___) || (___) || (___) || (____/\\    \r";
    vga.print_string2(rogue1);
    rogue1 = "    |/   \\__/(_______)(_______)(_______)(_______/    \r";
    vga.print_string2(rogue1);

    vga.setforecolor(vga.Yellow);
    vga.setbackcolor(vga.DeepFir);
    vga.row = 10;
    vga.col = 100;
    vga.print_string2("Moves:  \r");
    vga.print_string2(" y k u  \r");
    vga.print_string2(" h . l  \r");
    vga.print_string2(" b j n  \r");

    //vga.setforecolor(vga.Yellow);
    //vga.setbackcolor(vga.DeepFir);
    vga.row = 20;
    vga.col = 100;
    vga.print_string2("Actions:  \r");
    vga.print_string2("? help           \r");
    vga.print_string2("s search         \r");
    vga.print_string2("> down stairs    \r");
    vga.print_string2("< up stairs      \r");
    vga.print_string2("i inventory      \r");
    vga.print_string2("q quaff potion   \r");
    vga.print_string2("r read scroll    \r");
    vga.print_string2("e eat            \r");
    vga.print_string2("w wield weapon   \r");
    vga.print_string2("W wear armor     \r");
    vga.print_string2("T take off armor \r");
    vga.print_string2("P put on ring    \r");
    vga.print_string2("R remove ring    \r");
    vga.print_string2("t throw something\r");
    vga.print_string2("z zap a wand     \r");
    vga.print_string2("^ identify trap  \r");
    vga.print_string2(". rest a turn    \r");
    vga.print_string2("S save game      \r");
    vga.print_string2("@ print stats    \r");
    vga.print_string2("d drop object    \r");
    

    //vga.setforecolor(vga.Yellow);
    //vga.setbackcolor(vga.DeepFir);
    vga.row = 10;
    vga.col = 1;
    vga.print_string2("Monsters:      \r");
    vga.print_string2("A Aquator      \r");
    vga.print_string2("B Bat          \r");
    vga.print_string2("C Centaur      \r");
    vga.print_string2("D Dragon       \r");
    vga.print_string2("E Emu          \r");
    vga.print_string2("V Venus Flytrap\r");
    vga.print_string2("G Griffin      \r");
    vga.print_string2("H Hobgoblin    \r");
    vga.print_string2("I Ice monster  \r");
    vga.print_string2("J Jabberwock   \r");
    vga.print_string2("K Kestrel      \r");
    vga.print_string2("L Leprechaun   \r");
    vga.print_string2("M Medusa       \r");
    vga.print_string2("N Nymph        \r");
    vga.print_string2("O Orc          \r");
    vga.print_string2("P Phantom      \r");
    vga.print_string2("Q Quagga       \r");
    vga.print_string2("R Rattlesnake  \r");
    vga.print_string2("S Snake        \r");
    vga.print_string2("T Troll        \r");
    vga.print_string2("U Ur-vile      \r");
    vga.print_string2("V Vampire      \r");
    vga.print_string2("W Wraith       \r");
    vga.print_string2("X Xeroc        \r");
    vga.print_string2("Y Yeti         \r");
    vga.print_string2("Z Zombie       \r");

    //Start up file system for debugging
#if defined(USE_SD)
    mount("/files", _vfs_open_sdcard());
    chdir("/files");
#elif defined(USE_HOST)
    mount("/files", _vfs_open_host());
    chdir("/files");
#else
    #warning no file system specified, I hope you are compiling for PC
#endif


    //for (;;)
    //{
    //    vga.setforecolor(vga.white);
    //    vga.row = 2;
    //    vga.col = 2;
    //    vga.dec(MouseStatus.x);
    //    vga.print_string(" ");
    //    //char c = kbm.rawKey();
    //    //if (c == 0) 
    //    //{
    //    //    c = ' ';
    //    //}
    //    //vga.row = 4;
    //    //vga.col = 2;
    //    //vga.Print(c);
    //    //usleep(100000);
    //    vga.hex(MouseStatus.scan0,2);
    //    vga.print_string(" ");
    //    vga.hex(MouseStatus.scan1, 2);
    //    vga.print_string(" ");
    //    vga.hex(MouseStatus.scan2, 2);
    //    vga.print_string(" ");
    //    vga.hex(MouseStatus.scan3, 2);
    //    vga.print_string(" ");
    //    vga.hex(MouseStatus.scan4, 2);
    //    vga.print_string(" ");
    //    vga.hex(MouseStatus.scan5, 2);
    //    vga.print_string(" ");
    //    vga.hex(MouseStatus.scan6, 2);
    //    vga.print_string(" ");
    //    vga.hex(MouseStatus.scan7, 2);
    //    vga.print_string("   ");       
    //}

}

char MyGetChar()
{//RJA:  Get keyboard input from some source
#ifdef USE_USB_KEYBOARD
    char c;
    do
    {
        c = kbm.rawKey() & 0xFF;  //use P2 connected USB keyboard
    } while (c == 0);
    kbm.flushkeys();  //empty keyboard buffer    
    return SwitchChar(c);
#endif // USE_USB_KEYBOARD
    char c = getchar();  //use PC terminal keyboard
    SwitchChar(c);
    return c;
}


char MyGetCharSwitched()  //with character switching for keypad
{//RJA:  Get keyboard input from some source
#ifdef USE_USB_KEYBOARD
    char c;
    do
    {
        c=kbm.rawKey() & 0xFF;  //use P2 connected USB keyboard
    } while (c == 0);
    kbm.flushkeys();  //empty keyboard buffer    
    return SwitchChar(c);
#endif // USE_USB_KEYBOARD
    char c=getchar();  //use PC terminal keyboard   
    return SwitchChar(c);
}

char SwitchChar(char c)
{//replace keypad letters with numbers
        //RJA adding keypad numbers
    switch(c)
    {
    case '4': return 'h';
    case '2': return 'j';
    case '8': return 'k';
    case '6': return 'l';
    case '7': return 'y';
    case '9': return 'u';
    case '1': return 'b';
    case '3': return 'n';
    case '5': return 's';
    default: return c;
    }
}





void flush_type()
{//RJA:  not sure what this is...  seems to be about flushing input...

}

/*** Video attribute macros ***/
#define A_NORMAL      (chtype)0
# define A_OVERLINE   A_NORMAL
# define A_STRIKEOUT  A_NORMAL
# define A_ATTRIBUTES (chtype)0xffff0000
# define A_ALTCHARSET (chtype)0x00010000
# define A_RIGHTLINE  (chtype)0x00020000
# define A_LEFTLINE   (chtype)0x00040000
# define A_INVIS      (chtype)0x00080000
# define A_UNDERLINE  (chtype)0x00100000
# define A_REVERSE    (chtype)0x00200000
# define A_BLINK      (chtype)0x00400000
# define A_BOLD       (chtype)0x00800000
# define A_CHARTEXT   (chtype)0x0000ffff
# define A_COLOR      (chtype)0xff000000
# define A_RGB_COLOR  A_NORMAL
# define A_DIM        A_NORMAL
# define PDC_COLOR_SHIFT 24

typedef struct __window WINDOW;
typedef uint32_t chtype;

//Coordinate data type
typedef struct 
{
    int x;
    int y;
}Coord;

//namespace
//{
    const int s_baudrate = 3000;
    void * s_screen = 0; //DisplayInterface* s_screen = 0;
    void * s_input = 0; //InputInterface* s_input = 0;
//}

typedef chtype attr_t;

#define COLOR_PAIR(n)      (((chtype)(n) << PDC_COLOR_SHIFT) & A_COLOR)
#define PAIR_NUMBER(n)     ((((n) & A_COLOR) >> PDC_COLOR_SHIFT) & 0xff)

typedef struct 
{
    int Left;
    int Top;
    int Right;
    int Bottom;

    int Width() { return Right - Left + 1; }
    int Height() { return Bottom - Top + 1; }
}Region;

#define min(x,y)  ((x) > (y) ? (y) : (x))
#define max(x,y)  ((x) > (y) ? (x) : (y))


/*----------------------------------------------------------------------
 *
 *  PDCurses Manifest Constants
 *
 */

#ifndef FALSE
# define FALSE 0
#endif
#ifndef TRUE
# define TRUE 1
#endif
#ifndef NULL
# define NULL (void *)0
#endif
#ifndef ERR
# define ERR (-1)
#endif
#ifndef OK
# define OK 0
#endif

WINDOW* stdscr = 0;
WINDOW* curscr = 0;
WINDOW* hw = NULL;			/* used as a scratch window */
WINDOW* tw, * sw;  //RJA moving these here from things.c to avoid running out of memory

//
//int COLS = 0;
//int LINES = 0;

#define COLS 80
#define LINES 25

/* These will _only_ work as macros */

#define getbegyx(w, y, x)  (y = getbegy(w), x = getbegx(w))
#define getmaxyx(w, y, x)  (y = getmaxy(w), x = getmaxx(w))
#define getparyx(w, y, x)  (y = getpary(w), x = getparx(w))
#define getyx(w, y, x)     (y = getcury(w), x = getcurx(w))

#define getsyx(y, x)       { if (curscr->_leaveit) (y)=(x)=-1; \
                             else getyx(curscr,(y),(x)); }

#define CLEAR_MSG                     msg("")

//RJA:  Stuff from curses.cpp, useful here because can't create that WINDOW class...

int	standend(void)
{
    return wstandend(stdscr);
}

int	standout(void)
{
    return wstandout(stdscr);
}

int wstandend(WINDOW* w)
{
    return wattroff(w, COLOR_PAIR(0x70));
}

int wstandout(WINDOW* w)
{
    return wattron(w, COLOR_PAIR(0x70));
}

int mvaddstr(int r, int c, const char* s)
{
    return mvwaddstr(stdscr, r, c, s);
}

int mvwaddstr(WINDOW* w, int r, int c, const char* s)
{
    wmove(w, r, c);
    return waddstr(w, s);
}

int refresh(void)
{
    return wrefresh(stdscr);
}

int clear(void)
{
    return wclear(stdscr);
}

int	addch(chtype ch)
{
    return waddch(stdscr, ch);
}

int	addrawch(chtype ch)
{
    if (stdscr->parent != 0)
    {
        printf("bad parent addrawch\n");
        for (;;) {}
    }
    return waddrawch(stdscr, ch);
}

int addstr(const char* s)
{
    return waddstr(stdscr, s);
}

int attroff(chtype ch)
{
    return wattroff(stdscr, ch);
}

int attron(chtype ch)
{
    return wattron(stdscr, ch);
}

int attrset(chtype a)
{
    return wattrset(stdscr, a);
}

int endwin(void)
{
    return OK;
}

int	move(int r, int c)
{
    return wmove(stdscr, r, c);
}


int mvcur(int a, int b, int c, int d)
{
    return OK;
}

chtype mvinch(int r, int c)
{
    return mvwinch(stdscr, r, c);
}

chtype mvwinch(WINDOW* w, int r, int c)
{
    wmove(w, r, c);
    return winch(w);
}

int	mvaddch(int r, int c, chtype ch)
{
    char d = (char) ch;
    //printf("Adding char %c=%d\n", d,d);
    return mvwaddch(stdscr, r, c, ch);
}

int	mvaddrawch(int r, int c, chtype ch)
{
    return mvwaddrawch(stdscr, r, c, ch);
}

int mvwaddch(WINDOW* w, int r, int c, chtype ch)
{
    wmove(w, r, c);
    char d = (char)ch;
    //printf("Moved, now adding char %c=%d\n", d, d);
    return waddch(w, ch);
}

int mvwaddrawch(WINDOW* w, int r, int c, chtype ch)
{
    wmove(w, r, c);
    return waddrawch(w, ch);
}

int	clrtoeol(void)
{
    return wclrtoeol(stdscr);
}

int getnstr(char* dest, int n)
{
    return wgetnstr(stdscr, dest, n);
}

chtype inch(void)
{
    return winch(stdscr);
}

//typedef char* va_list;
typedef uint32_t va_list;

int	mvprintw(int r, int c, const char* f, ...)
{
    char buf[1024 * 16];
    va_list argptr;
    va_start(argptr, f);
    vsprintf(buf, f, argptr);
    va_end(argptr);

    return mvwaddstr(stdscr, r, c, buf);
}

int wprintw(WINDOW* w, const char* f, ...)
{
    char buf[1024 * 16];
    va_list argptr;
    va_start(argptr, f);
    vsprintf(buf, f, argptr);
    va_end(argptr);

    return waddstr(w, buf);
}

int mvwprintw(WINDOW* w, int r, int c, const char* f, ...)
{
    char buf[1024 * 16];
    va_list argptr;
    va_start(argptr, f);
    vsprintf(buf, f, argptr);
    va_end(argptr);

    wmove(w, r, c);
    return waddstr(w, buf);
}

int printw(const char* f, ...)
{
    char buf[1024 * 16];
    va_list argptr;
    va_start(argptr, f);
    vsprintf(buf, f, argptr);
    va_end(argptr);

    return waddstr(stdscr, buf);
}

char* unctrl(chtype ch)
{
    static char chstr[9];

    if (isspace(ch))
        strcpy(chstr, " ");
    else if (!isprint(ch))
        if (ch < ' ')
            sprintf(chstr, "^%c", ch + '@');
        else
            sprintf(chstr, "\\x%x", ch);
    else {
        chstr[0] = (unsigned char)ch;
        chstr[1] = 0;
    }
    return chstr;
}

int leaveok(WINDOW* w, bool b)
{
    return OK;
}

int touchwin(WINDOW* w)
{
    wfill_dirty(w,1);
    return OK;
}

int clearok(WINDOW* w, bool enable)
{
    w->should_clear_screen = enable;
    return OK;
}

int mvwin(WINDOW* w, int r, int c)
{
    w->origin.x = c;// = { c, r };
    w->origin.y = r;
    return OK;
}




void* mallocsafe(int n)
{
    //printf("allocating %d bytes, press any key to continute\n", n);
    //MyGetChar();
    void* p = _gc_alloc_managed(n);
    if (p == 0)
    {
        
        printf("malloc failed to allocate %d bytes\n, press any key to try garbage collection...", n);
        MyGetChar();
        _gc_collect();
        p = _gc_alloc_managed(n);
        if (p==0)
            for (;;) {}
        printf("garbage collection worked!, press any key..");
        MyGetChar();
    }
    return p;
}

void* callocsafe(int n, int size)
{
    n = n * size;
    //printf("allocating %d bytes, press any key to continute\n", n);
    //MyGetChar();  
    _gc_collect();
    void* p = calloc(n,1);
    if (p == 0)
    {

        printf("calloc failed to allocate %d bytes\n, press any key to try garbage collection again...", n);
        MyGetChar();
        _gc_collect();
        p = calloc(n,1);
        if (p == 0)
        {
            printf("calloc(%d,1) failed again\n", n);
            for (;;) {}
        }
        printf("garbage collection worked!, press any key..");
        MyGetChar();
    }
    return p;
}



WINDOW* newwin(int nlines, int ncols, int begin_y, int begin_x)
{
    //return new WINDOW(nlines, ncols, begin_y, begin_x);
    //printf("creating new scratch window, press a key...\n");
    //MyGetChar();
    WINDOW* w = mallocsafe(sizeof(WINDOW));
    //scratch = w;
    wCreate(w, nlines, ncols, begin_y, begin_x);
    return w;// scratch;
}

WINDOW* subwin(WINDOW* p, int nlines, int ncols, int begin_y, int begin_x)
{
    //return new WINDOW(w, nlines, ncols, begin_y, begin_x);
    //printf("creating sub-window with %d lines and %d cols, press a key...\n",nlines,ncols);
    //MyGetChar();
    WINDOW* w = mallocsafe(sizeof(WINDOW));
    //scratchsub = w;
    wCreateChild(w, p, nlines, ncols, begin_y, begin_x);
    //w->parent = p;
    return w;
}


WINDOW* initscr(void)
{
    //if (LINES == 0)
    //    LINES = 25;
    //if (COLS == 0)
    //    COLS = 80;

    //if (s_screen)
    //    s_screen->SetDimensions({ COLS, LINES });

    //stdscr = new __window(LINES, COLS, 0, 0);
    //curscr = new __window(LINES, COLS, 0, 0);

    WINDOW* w = mallocsafe(sizeof(WINDOW));
    stdscr = w;
    wCreate(stdscr, LINES, COLS, 0, 0);
    printf("Windex=%d, dimensions=%d,%d\n", windex(stdscr, 1, 1), stdscr->dimensions.x, stdscr->dimensions.y);
    w = mallocsafe(sizeof(WINDOW));
    curscr = w;
    wCreate(curscr, LINES, COLS, 0, 0);

    return stdscr;
}

typedef struct  __window
{//RJA creating FlexC compatible version of pdcurses window

    Coord origin;
    Coord dimensions;

    chtype *m_data;
    char *m_dirty;

    chtype attr;

    int row;
    int col;

    bool no_delay;
    bool should_echo;
    bool should_clear_screen;

    WINDOW* parent;

};

//RJA declaring these static as don't work when made with malloc
WINDOW *stdscr;
WINDOW *curscr;
//WINDOW *scratch; //for hw
//WINDOW* scratchsub; //for hw

void wCreate(WINDOW* w, int lines, int cols, int begin_y, int begin_x)
{
    printf("window size =%d\n", sizeof(WINDOW));
    //w = malloc(sizeof(WINDOW));
    w->origin.x = 0;// = Coord{ 0, 0 };
    w->origin.y = 0;
    w->dimensions.x = 0;// = { 0, 0 };
    w->dimensions.y = 0;
    w->m_data = 0;
    w->attr = 0;
    w->row = 0;
    w->col = 0;
    w->no_delay = false;
    w->should_echo = true;
    w->should_clear_screen = false;

    if (lines == 0)
        lines = LINES - begin_y;
    if (cols == 0)
        cols = COLS - begin_x;

    w->dimensions.x = cols;// = { cols, lines };
    w->dimensions.y = lines;
    w->origin.x = begin_x;// = { begin_x, begin_y };
    w->origin.y = begin_y;//

    w->m_data = mallocsafe(lines * cols * sizeof(chtype)); //chtype[lines * cols];
    w->m_dirty = mallocsafe(lines * cols * sizeof(char));
    werase(w);
    w->parent = 0;

}

void wCreateChild(WINDOW* w, WINDOW* p, int lines, int cols, int begin_y, int begin_x)
{
    w->dimensions.x = cols;// = { cols, lines };
    w->dimensions.y = lines;//
    w->origin.x = begin_x;// = { begin_x, begin_y };
    w->origin.y = begin_y;

    w->m_data = p->m_data;
    w->m_dirty = p->m_dirty;

    w->parent = p;
}

void wReset(WINDOW* w, int lines, int cols, int begin_y, int begin_x)
{//RJA adding this to make things windows static

    w->origin.x = 0;// = Coord{ 0, 0 };
    w->origin.y = 0;
    w->dimensions.x = 0;// = { 0, 0 };
    w->dimensions.y = 0;

    w->attr = 0;
    w->row = 0;
    w->col = 0;
    w->no_delay = false;
    w->should_echo = true;
    w->should_clear_screen = false;

    if (lines == 0)
        lines = LINES - begin_y;
    if (cols == 0)
        cols = COLS - begin_x;

    w->dimensions.x = cols;// = { cols, lines };
    w->dimensions.y = lines;
    w->origin.x = begin_x;// = { begin_x, begin_y };
    w->origin.y = begin_y;//

    werase(w);
    w->parent = 0;

}

void wResetChild(WINDOW* w, WINDOW* p, int lines, int cols, int begin_y, int begin_x)
{//RJA adding this to make things windows static
    w->dimensions.x = cols;// = { cols, lines };
    w->dimensions.y = lines;//
    w->origin.x = begin_x;// = { begin_x, begin_y };
    w->origin.y = begin_y;

    w->m_data = p->m_data;
    w->m_dirty = p->m_dirty;

    w->parent = p;
}


void wfill_dirty(WINDOW* w, int value)
{
    memset(w->m_dirty, value, w->dimensions.x * w->dimensions.y * sizeof(char));
}

int werase(WINDOW* w)
{
    wmove(w, 0, 0);
    for (int r = 0; r < w->dimensions.y; ++r)
        for (int c = 0; c < w->dimensions.x; ++c)
            wset_data(w, r, c, ' ');
    return OK;
}

int wmove(WINDOW* w, int r, int c)
{
    w->row = r;
    w->col = c;
    return OK;
}

int windex(WINDOW* w, int r, int c)
{
    return r * w->dimensions.x + c;
}

chtype wget_data(WINDOW* w, int r, int c)
{
    Coord o = wdata_coords(w, r, c);
    return *wdata(w, o.y, o.x);
}

chtype wget_data_absolute(WINDOW* w, int abs_r, int abs_c)
{
    return wget_data(w, abs_r - w->origin.y, abs_c - w->origin.x);
}

void wset_data(WINDOW* w, int r, int c, chtype ch)
{
    Coord o = wdata_coords(w, r, c);
    //char d = (char) ch&0xff;
    //if ((d>32)&&(d<'z'))
    //    printf("Setting data=%c at %d,%d\n", d, o.y, o.x);

    //if (w->parent != 0)
    //{
    //    chtype* p1 = wdata(w->parent, o.y, o.x);
    //    chtype* p2 = wdata(w, o.y, o.x);
    //    printf("setting data on parent %d=%d ?, press any key\n",p1,p2);
    //    MyGetChar();
    //    *wdata(w->parent, o.y, o.x) = ch;
    //    return;
    //}
    *wdata(w, o.y, o.x) = ch;
    //if (w->parent != 0)
    //{
    //    printf("bad parent wset_data\n");
    //    for (;;) {}
    //}
    *wdirty(w, o.y, o.x) = 1;
    //printf("dirty\n");
}

    //void set_data_absolute(int abs_r, int abs_c, chtype ch)
    //{
    //    set_data(abs_r - origin.y, abs_c - origin.x, ch);
    //}

chtype* wdata(WINDOW* w, int r, int c)
{
    return w->parent ? wdata(w->parent, r, c) : &w->m_data[windex(w, r, c)];
}

char* wdirty(WINDOW* w, int r, int c)
{
    //printf("dirty parent= %d\n", w->parent);
    return w->parent ? wdirty(w->parent, r, c) : &w->m_dirty[windex(w,r, c)];
}

Coord wdata_coords(WINDOW* w, int r, int c)
{
    Coord p = { c, r };
    if (w->parent) {
        p.x += w->origin.x - w->parent->origin.x;
        p.x += w->origin.y - w->parent->origin.y;
    }
    return p;
}


int waddch(WINDOW* w, chtype ch)
{
    ch |= w->attr;
    if ((ch & A_CHARTEXT) == '\n')
    {
        wclrtoeol(w);
        ++w->row;
        w->col = 0;
    }
    else if ((ch & A_CHARTEXT) == '\b')
    {
        if (w->col > 0)
            --w->col;
        wset_data(w, w->row, w->col, ' ');
    }
    else if ((ch & A_CHARTEXT) == '\t')
    {
        do {
            wset_data(w, w->row, w->col, ' ');
            ++w->col;
        } while (w->col % 8);
    }
    else {
        //char d = (char) ch;
        //printf("Setting data %c at %d, %d; attr=%d\n", d, w->row, w->col, w->attr);
        wset_data(w, w->row, w->col, ch);
        ++w->col;
    }
    return OK;
}

int wclrtoeol(WINDOW* w)
{
    for (int c = w->col; c < w->dimensions.x; ++c)
        wset_data(w, w->row, c, 0);
    return OK;
}

int waddrawch(WINDOW* w, chtype ch)
{
    ch |= w->attr;
    ch |= A_ALTCHARSET;
    if (w->parent != 0)
    {
        printf("bad parent waddrawch\n");
        for (;;) {}
    }
    wset_data(w, w->row, w->col, ch);
    ++w->col;
    return OK;
}

int waddstr(WINDOW* w, const char* s)
{
    while (*s)
        waddch(w,*s++);
    return OK;
}

int wattroff(WINDOW* w, chtype ch)
{
    if ((ch | A_COLOR) == (w->attr | A_COLOR))
    {
        w->attr &= ~A_COLOR; //clear the color
    }
    return OK;
}

int wattrset(WINDOW* w, chtype a)
{
    w->attr = (a & A_ATTRIBUTES);
    return OK;
}

int wattron(WINDOW* w, chtype ch)
{
    if ((ch | A_COLOR) && (w->attr | A_COLOR))
    {
        w->attr &= ~A_COLOR; //clear the old color
    }
    w->attr |= (ch & A_COLOR);
    return OK;
}

int wclear(WINDOW* w)
{
    werase(w);
    return OK;
}

int getcury(WINDOW* w)
{
    return w->row;
}

int getcurx(WINDOW* w)
{
    return w->col;
}

int getmaxy(WINDOW* w)
{
    return w->dimensions.y;
}

int getmaxx(WINDOW* w)
{
    return w->dimensions.x;
}

    //int getch()
    //{
    //    //if (!s_input)
    //    //    return ERR;
    //    int ch = MyGetChar();// s_input->MyGetChar(!no_delay, false, nullptr);
    //    //todo:mdk should consider echo here?
    //    return ch ? ch : ERR;
    //}


    //char* getsnstr_impl(unsigned int n)
    //{
    //    const int ESCAPE = 0x1b;

    //    char* s=new (n+1);
    //    while (true)
    //    {
    //        char c = MyGetChar();// s_input->MyGetChar(true, true, nullptr);
    //        switch (c)
    //        {
    //        case ESCAPE:
    //            //s.clear();
    //            //s.push_back(ESCAPE);
    //            s[0] = ESCAPE;
    //            s[1] = 0;
    //            return s;
    //        case '\b':
    //            if (!(strlen(s)==0)){//s.empty()) {
    //                if (should_echo) {
    //                    addch('\b');
    //                }
    //                strncat(s, &c, 1);    //s.pop_back();
    //            }
    //            break;
    //        default:
    //            if (strlen(s) >=n){ //s.size() >= n) {
    //                //beep();
    //                break;
    //            }
    //            if (should_echo) {
    //                addch(c);
    //            }
    //            strncat(s, &c, 1);    //s.pop_back();
    //            //s.push_back(c);
    //            break;
    //        case '\n':
    //        case '\r':
    //        case 0:
    //            return s;
    //        }

    //        refresh();
    //    }
    //}

int wrefresh(WINDOW* w)
{//RJA this appears to copy a window to the displayed window
    if (w->should_clear_screen || curscr->should_clear_screen)
    {
        wfill_dirty(w,1);
        wfill_dirty(curscr,1);
        w->should_clear_screen = false;
        curscr->should_clear_screen = false;
    }

    //RJA going to try erasing curscr instead of dealing with dirty
    werase(curscr);

    for (int r = 0; r < w->dimensions.y; ++r) {
        memcpy(wdata(curscr, r + w->origin.y, w->origin.x), wdata(w, r, 0), w->dimensions.x * sizeof(chtype));
        memcpy(wdirty(curscr, r + w->origin.y, w->origin.x), wdirty(w, r, 0), w->dimensions.x * sizeof(char));
    }

    if (s_screen) {
        //s_screen->UpdateRegion(curscr->m_data.get(), curscr->m_dirty.get());
        //s_screen->MoveCursor({ col + origin.x, row + origin.y });
    }
    DrawScreen();
    wfill_dirty(w,0);
    wfill_dirty(curscr,0);


    return OK;
}

    //int getnstr(char* dest, int n)
    //{
    //    if (!s_input)
    //        return ERR;

    //    char* s = getsnstr_impl(n - 1);
    //    strcpy(dest, s);
    //    return OK;
    //}

    //int chgat(int n, attr_t attr, short color, const void* opts)
    //{
    //    if (n == -1)
    //        n = COLS - col;

    //    for (int c = col; c < col + n; ++c) {
    //        chtype ch = ((get_data(row, c) & A_CHARTEXT) | attr | COLOR_PAIR(color));
    //        set_data(row, c, ch);
    //    }

    //    return OK;
    //}

    chtype winch(WINDOW *w)
    {
        //todo: should i return attrs?
        return wget_data(w, w->row, w->col);
    }

    //int move(int r, int c)
    //{
    //    row = r;
    //    col = c;
    //    return OK;
    //}

    //int mvwin(int r, int c)
    //{
    //    origin.x = c;// = { c, r };
    //    origin.y = r;
    //    return OK;
    //}

    //int nodelay(bool enable)
    //{
    //    no_delay = enable;
    //    return OK;
    //}

    //int echo()
    //{
    //    should_echo = true;
    //    return OK;
    //}

    //int noecho()
    //{
    //    should_echo = false;
    //    return OK;
    //}

    //int touchwin()
    //{
    //    fill_dirty(1);
    //    return OK;
    //}

    //int clearok(bool enable)
    //{
    //    should_clear_screen = enable;
    //    return OK;
    //}

    //void fill_dirty(int value)
    //{
    //    //memset(m_dirty.get(), value, dimensions.x * dimensions.y * sizeof(char));
    //    memset(m_dirty, value, dimensions.x * dimensions.y * sizeof(char));
    //}

    //int overlay(WINDOW* dest, bool copy_spaces) 
    //{
    //    Region r1 = window_region();
    //    Region r2 = dest->window_region();

    //    // calculate the overlap of the windows in absolute screen coordinates
    //    Region overlap;
    //    overlap.Left = max(r1.Left, r2.Left);
    //    overlap.Top = max(r1.Top, r2.Top);
    //    overlap.Right = min(r1.Right, r2.Right);
    //    overlap.Bottom = min(r1.Bottom, r2.Bottom);

    //    for (int r = overlap.Top; r <= overlap.Bottom; ++r) {
    //        for (int c = overlap.Left; c <= overlap.Right; ++c) {
    //            ch = get_data_absolute(r, c);
    //            if (copy_spaces || (ch & A_CHARTEXT) != ' ')
    //                dest->set_data_absolute(r, c, ch);
    //        }
    //    }
    //    return OK;
    //}


    //Region window_region() 
    //{
    //    Region r;
    //    r.Left = origin.x;
    //    r.Top = origin.y;
    //    r.Right = r.Left + dimensions.x - 1;
    //    r.Bottom = r.Top + dimensions.y - 1;
    //    return r;
    //}

    //int copywin(WINDOW* dest, int srcrow, int srccol, int destrow, int destcol, int destmaxrow, int destmaxcol)
    //{
    //    int roffset = srcrow - destrow;
    //    int coffset = srccol - destcol;
    //    int ncols = destmaxcol - destcol + 1;
    //    for (int r = destrow; r <= destmaxrow; ++r) {
    //        int c = destcol;
    //        memcpy(dest->data(r, c), this->data(r + roffset, c + coffset), ncols * sizeof(chtype));
    //        memset(dest->dirty(r, c), 1, ncols * sizeof(char));
    //    }
    //    return OK;
    //}

    //int index(int r, int c) 
    //{
    //    return r * dimensions.x + c;
    //}

    //chtype get_data(int r, int c)
    //{
    //    Coord o = data_coords(r, c);
    //    return *data(o.y, o.x);
    //}

    //chtype get_data_absolute(int abs_r, int abs_c)
    //{
    //    return get_data(abs_r - origin.y, abs_c - origin.x);
    //}

    //void set_data(int r, int c, chtype ch)
    //{
    //    Coord o = data_coords(r, c);
    //    *data(o.y, o.x) = ch;
    //    *dirty(o.y, o.x) = 1;
    //}

    //void set_data_absolute(int abs_r, int abs_c, chtype ch)
    //{
    //    set_data(abs_r - origin.y, abs_c - origin.x, ch);
    //}

    ////chtype* data(int r, int c)
    ////{
    ////    return parent ? parent.data(r, c) : &m_data[index(r, c)];
    ////}

    //char* dirty(int r, int c)
    //{
    //    return parent ? parent.dirty(r, c) : &m_dirty[index(r, c)];
    //}

    //Coord data_coords(int r, int c)
    //{
    //    Coord p = { c, r };
    //    if (parent) {
    //        p.x += origin.x - parent.origin.x;
    //        p.x += origin.y - parent.origin.y;
    //    }
    //    return p;
    //}

    ////__window* parent = 0;
    //void* parent;



int flushinp(void)
{
    //if (s_input)
    //    s_input->Flush();
    return OK;
}

char erasechar(void)
{
    return '\b';
}


int	delwin(WINDOW* w)
{
    //delete w;
    if (w->parent == 0)
    {//top level windows have data and dirtly, child windows do not
        free(w->m_data);
        free(w->m_dirty);
    }
    free(w);
    if (w == curscr || w == stdscr) 
    {//RJA doesn't think we should ever get here...
        curscr = 0;
        stdscr = 0;
    }
    return OK;
}

//////////////////////////////////////////////
//from mdport.c
int
md_hasclreol(void)
{
    return(TRUE);

}

//////////////////////////////////////////////
//code that draws screen on FlexCC terminal window


//Terminal escape sequenses
#define ESC "\x1b"
#define CSI "\x1b["

#define VgaOffsetX 18
#define VgaOffsetY 17

void DrawScreen()
{//draw current screen on terminal and/or VGA
    WINDOW* w = curscr;

    //printf("Drawing screen of size %d, %d.\n\n", LINES, COLS);
    printf(CSI "2J"); // Clear screen
    printf(CSI "1;1H"); // Home

    vga.SetForeColor(vga.White);
    vga.SetBackColor(vga.Black);

    for (int j = 0; j < LINES; j++)
    {
        vga.SetRow(j*2+ VgaOffsetY);
        vga.SetCol(VgaOffsetX);
        char b = '0';
        for (int i = 0; i < COLS; i++)
        {

            //char c = (char)(chat(j, i) & A_CHARTEXT);// wdirty(w, j, i);
            chtype c = *wdata(w, j, i);// (stdscr, j, i);
            char d = (char)c;
            if (d < ' ')
            {
                d = ' ';
            }
#ifdef DISPLAY_IN_TERMINAL
            putchar(d);
#endif // DISPLAY_IN_TERMINAL
            vga.Print(d);

        }
#ifdef DISPLAY_IN_TERMINAL
        putchar('\n');
#endif // DISPLAY_IN_TERMINAL
    }

}

void DumpScreen()
{//save screen to file on uSD for debugging
    WINDOW* w = curscr;
    FILE* f;
    f = fopen("rogue.txt", "w");

    if (!f) {
        printf("Unable to open rogue.txt\n");
        for (;;) {}
    }

    for (int j = 0; j < LINES; j++)
    {
        for (int i = 0; i < COLS; i++)
        {
            chtype c = *wdata(w, j, i);
            char d = (char)c;
            fputc(d, f);
        }
        fputc('\n', f);
    }
    fclose(f);
    printf("Screen saved to file rogue.dat, press any key\n");
    MyGetChar();
}



#endif // !CURSES

